"""
This module contains classes and methods for working with xyz unevenly sensors.
"""

import typing

import numpy

import redvox.api900.constants as constants
from redvox.api900.sensors.unevenly_sampled_channel import UnevenlySampledChannel
from redvox.api900.sensors.unevenly_sampled_sensor import UnevenlySampledSensor


class XyzUnevenlySampledSensor(UnevenlySampledSensor):
    """
    This class subclasses the UnevenlySampledSensor class and provides methods for working with channels that provide
    data in the X, Y, and Z dimensions.
    """

    def __init__(self, unevenly_sampled_channel: UnevenlySampledChannel, x_type: int, y_type: int, z_type: int):
        """
        Initializes this class.
        :param unevenly_sampled_channel: An instance of an UnevenlySampledChannel.
        :param x_type: The X channel type enum.
        :param y_type: The Y channel type enum.
        :param z_type: The Z channel type enum.
        """
        super().__init__(unevenly_sampled_channel)
        self._unevenly_sampled_channel.set_channel_types([x_type, y_type, z_type])
        self._x_type = x_type
        self._y_type = y_type
        self._z_type = z_type

    def _payload_values(self) -> numpy.ndarray:
        """
        Returns this channel's payload as an interleaved payload of the form
        [[x_0, y_0, z_0], [x_1, y_1, z_1], ..., [x_n, y_n, z_n]].
        :return: This channel's payload as an interleaved payload.
        """
        return self._unevenly_sampled_channel.get_multi_payload([
            self._x_type,
            self._y_type,
            self._z_type
        ])

    def _set_payload_values(self,
                            x_values: typing.Union[typing.List, numpy.ndarray],
                            y_values: typing.Union[typing.List, numpy.ndarray],
                            z_values: typing.Union[typing.List, numpy.ndarray],
                            pl_type: constants.PayloadType) -> 'XyzUnevenlySampledSensor':
        """
        Sets the interleaved payload of this XYZ channel given the X, Y, Z values and payload type.
        :param x_values: The x values.
        :param y_values: The y values.
        :param z_values: The z values.
        :param pl_type: Payload type.
        :return: An instance of the sensor.
        """
        self._unevenly_sampled_channel.set_interleaved_payload([
            x_values,
            y_values,
            z_values
        ],
                                                               pl_type)
        return self

    def payload_values_x(self) -> numpy.ndarray:
        """
        Returns the x-component of this channel's payload.
        :return: The x-component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(self._x_type)

    def payload_values_y(self) -> numpy.ndarray:
        """
        Returns the y-component of this channel's payload.
        :return: The y-component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(self._y_type)

    def payload_values_z(self) -> numpy.ndarray:
        """
        Returns the z-component of this channel's payload.
        :return: The z-component of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_payload(self._z_type)

    def payload_values_x_mean(self) -> float:
        """
        Returns the x-component mean of this channel's payload.
        :return: The x-component mean of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(self._x_type)

    def payload_values_y_mean(self) -> float:
        """
        Returns the y-component mean of this channel's payload.
        :return: The y-component mean of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(self._y_type)

    def payload_values_z_mean(self) -> float:
        """
        Returns the z-component mean of this channel's payload.
        :return: The z-component mean of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_mean(self._z_type)

    def payload_values_x_median(self) -> float:
        """
        Returns the x-component median of this channel's payload.
        :return: The x-component median of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(self._x_type)

    def payload_values_y_median(self) -> float:
        """
        Returns the y-component median of this channel's payload.
        :return: The y-component median of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(self._y_type)

    def payload_values_z_median(self) -> float:
        """
        Returns the z-component median of this channel's payload.
        :return: The z-component median of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_median(self._z_type)

    def payload_values_x_std(self) -> float:
        """
        Returns the x-component standard deviation of this channel's payload.
        :return: The x-component standard deviation of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(self._x_type)

    def payload_values_y_std(self) -> float:
        """
        Returns the y-component standard deviation of this channel's payload.
        :return: The y-component standard deviation of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(self._y_type)

    def payload_values_z_std(self) -> float:
        """
        Returns the z-component standard deviation of this channel's payload.
        :return: The z-component standard deviation of this channel's payload.
        """
        return self._unevenly_sampled_channel.get_value_std(self._z_type)



"""
Provides library level metadata and constants.
"""

NAME = "redvox"
VERSION = "3.0.0a14"


def version() -> str:
    """Returns the version number of this library."""
    return VERSION


def print_version():
    """Prints the version number of this library"""
    print(version())



"""
This module provides a high level API for creating, reading, and editing RedVox compliant API 1000 files.
"""

from datetime import datetime, timedelta
import os.path
from functools import total_ordering
from typing import Optional, List
from google.protobuf import json_format

import redvox.api1000.common.lz4
import redvox.api1000.common.typing
import redvox.api1000.errors as errors
import redvox.api1000.wrapped_redvox_packet.sensors.sensors as _sensors
import redvox.api1000.wrapped_redvox_packet.station_information as _station_information
import redvox.api1000.wrapped_redvox_packet.timing_information as _timing_information
import redvox.common.date_time_utils as dt_utils

from redvox.api1000.common.generic import ProtoBase, ProtoRepeatedMessage
from redvox.api1000.proto.redvox_api_m_pb2 import RedvoxPacketM
from redvox.api1000.wrapped_redvox_packet.event_streams import EventStream


@total_ordering
class WrappedRedvoxPacketM(ProtoBase[RedvoxPacketM]):
    """
    Wraps a RedVox API M protobuf buffer.
    """
    def __init__(self, redvox_proto: RedvoxPacketM):
        super().__init__(redvox_proto)

        self._station_information: _station_information.StationInformation = _station_information.StationInformation(
            redvox_proto.station_information)

        self._timing_information: _timing_information.TimingInformation = _timing_information.TimingInformation(
            redvox_proto.timing_information)

        self._sensors: _sensors.Sensors = _sensors.Sensors(redvox_proto.sensors)

        self._event_streams: ProtoRepeatedMessage = ProtoRepeatedMessage(
            redvox_proto,
            redvox_proto.event_streams,
            "event_streams",
            lambda event_stream_proto: EventStream(event_stream_proto),
            lambda event_stream: event_stream.get_proto()
        )

    # Implement methods required for total_ordering
    def __eq__(self, other) -> bool:
        """
        Tests if this packet is equal in time to another packet by comparing start mach timestamps.
        :param other: Other packet to compare against
        :return: True if the packets mach timestamps match, False otherwise
        """
        self_ts: float = self.get_timing_information().get_packet_start_mach_timestamp()
        other_ts: float = other.get_timing_information().get_packet_start_mach_timestamp()
        return self_ts == other_ts

    def __lt__(self, other: 'WrappedRedvoxPacketM') -> bool:
        """
        Checks if this packet is less than another packet by comparing start mach times.
        :param other: Other packet to compare against.
        :return: True if this packet is less than the other packet
        """
        self_ts: float = self.get_timing_information().get_packet_start_mach_timestamp()
        other_ts: float = other.get_timing_information().get_packet_start_mach_timestamp()
        return self_ts < other_ts

    @staticmethod
    def new() -> 'WrappedRedvoxPacketM':
        """
        Returns a new default instance of a WrappedRedvoxPacketApi1000.
        :return: A new default instance of a WrappedRedvoxPacketApi1000.
        """
        return WrappedRedvoxPacketM(RedvoxPacketM())

    @staticmethod
    def from_compressed_bytes(data: bytes) -> 'WrappedRedvoxPacketM':
        """
        Deserializes the byte content of an API M encoded .rdvxz file.
        :param data: The compressed bytes to deserialize.
        :return: An instance of a WrappedRedvoxPacketAPi1000.
        """
        redvox.api1000.common.typing.check_type(data, [bytes])
        uncompressed_data: bytes = redvox.api1000.common.lz4.decompress(data)
        proto: RedvoxPacketM = RedvoxPacketM()
        proto.ParseFromString(uncompressed_data)
        return WrappedRedvoxPacketM(proto)

    @staticmethod
    def from_compressed_path(rdvxm_path: str) -> 'WrappedRedvoxPacketM':
        """
        Deserialize an API M encoded .rdvxm file from the specified file system path.
        :param rdvxm_path: Path to the API M encoded file.
        :return: An instance of a WrappedRedvoxPacketApiM.
        """
        redvox.api1000.common.typing.check_type(rdvxm_path, [str])

        if not os.path.isfile(rdvxm_path):
            raise errors.WrappedRedvoxPacketMError(f"Path to file={rdvxm_path} does not exist.")

        with open(rdvxm_path, "rb") as rdvxz_in:
            compressed_bytes: bytes = rdvxz_in.read()
            return WrappedRedvoxPacketM.from_compressed_bytes(compressed_bytes)

    @staticmethod
    def from_json(json_str: str) -> 'WrappedRedvoxPacketM':
        """
        read json packet representing an API 1000 packet
        :param json_str: contains the json representing the packet
        :return: An instance of a WrappedRedvoxPacketM
        """
        return WrappedRedvoxPacketM(json_format.Parse(json_str, RedvoxPacketM()))

    @staticmethod
    def from_json_path(json_path: str) -> 'WrappedRedvoxPacketM':
        """
        read json from a file representing an api 1000 packet
        :param json_path: the path to the file to read
        :return: wrapped redvox packet api 1000
        """
        with open(json_path, "r") as json_in:
            return WrappedRedvoxPacketM.from_json(json_in.read())

    def default_filename(self, extension: Optional[str] = "rdvxm") -> str:
        """
        Returns the default filename for a given packet.
        :param extension: An (optional) file extension to add to the default file name.
        :return: The default filename for this packet.
        """
        # Format to be exactly 10 characters
        station_id: str = f"{self.get_station_information().get_id():0>10}"
        timestamp: int = round(self.get_timing_information().get_packet_start_mach_timestamp())
        filename: str = f"{station_id}_{timestamp}"

        if extension is not None:
            filename = f"{filename}.{extension}"

        return filename

    def default_file_dir(self) -> str:
        """
        Computes the default file directory structure for a structured layout for this particular packet.
        :return:  Directory structure for this packet when using structured layout
        """
        ts: float = self.get_timing_information().get_packet_start_mach_timestamp()
        dt: datetime = dt_utils.datetime_from_epoch_microseconds_utc(ts)
        year: str = f"{dt.year}:0>4"
        month: str = f"{dt.month:0>2}"
        day: str = f"{dt.day:0>2}"
        hour: str = f"{dt.hour:0>2}"
        return os.path.join(year, month, day, hour)

    def default_file_path(self) -> str:
        """
        Computes the default directory structure and file name for this packet for structured layouts.
        :return: The default directory structure and file name for this packet for structured layouts.
        """
        return os.path.join(self.default_file_dir(), self.default_filename())

    def write_compressed_to_file(self, base_dir: str, filename: Optional[str] = None) -> str:
        """
        Writes this packet to a .rdvxm file.
        :param base_dir: Directory to write .rdvxm to.
        :param filename: The (optional) file name to use. Will use default file name otherwise.
        :return: Path of written file.
        """
        if filename is None:
            filename = self.default_filename("rdvxm")

        if not os.path.isdir(base_dir):
            raise errors.WrappedRedvoxPacketMError(f"Base directory={base_dir} does not exist.")

        out_path: str = os.path.join(base_dir, filename)
        with open(out_path, "wb") as compressed_out:
            compressed_out.write(self.as_compressed_bytes())

        return out_path

    def write_json_to_file(self, base_dir: str, filename: Optional[str] = None) -> str:
        """
        Writes this packet to a .json file.
        :param base_dir: Directory to write .json to.
        :param filename: The (optional) file name to use. Will use default file name otherwise.
        :return: Path of written file.
        """
        if filename is None:
            filename = self.default_filename("json")

        if not os.path.isdir(base_dir):
            raise errors.WrappedRedvoxPacketMError(f"Base directory={base_dir} does not exist.")

        out_path: str = os.path.join(base_dir, filename)
        with open(out_path, "w") as json_out:
            json_out.write(self.as_json())

        return out_path

    def validate(self) -> List[str]:
        """
        Validates this packet.
        :return: A list of validation errors.
        """
        return validate_wrapped_packet(self)

    # Top-level packet fields
    def get_api(self) -> float:
        """
        Returns the API version of this packet.
        :return: The API version of this packet.
        """
        return self._proto.api

    def set_api(self, api: float) -> 'WrappedRedvoxPacketM':
        """
        Sets the api version of this packet.
        :param api: The API version (should be 1000.0)
        :return: The modified instance of this packet
        """
        redvox.api1000.common.typing.check_type(api, [int, float])
        self._proto.api = api
        return self

    def get_sub_api(self) -> float:
        """
        Returns the sub_api version. This version tracks small changes within API 1000.
        :return: The sub_api version. This version tracks small changes within API 1000.
        """
        return self._proto.sub_api

    def set_sub_api(self, sub_api: float) -> 'WrappedRedvoxPacketM':
        """
        Sets the sub_api.
        :param sub_api: sub_api to set.
        :return: A modified instance of self.
        """
        redvox.api1000.common.typing.check_type(sub_api, [int, float])
        self._proto.sub_api = sub_api
        return self

    def get_station_information(self) -> _station_information.StationInformation:
        """
        :return: An instance of StationInformation
        """
        return self._station_information

    def get_timing_information(self) -> _timing_information.TimingInformation:
        """
        :return: An instance of TimingInformation
        """
        return self._timing_information

    def get_sensors(self) -> _sensors.Sensors:
        """
        :return: An instance of Sensors
        """
        return self._sensors

    def get_event_streams(self) -> ProtoRepeatedMessage:
        """
        :return: A collection of event streams.
        """
        return self._event_streams

    # todo: add packet_duration calculations that don't rely on sensors existing if possible

    def get_packet_duration(self) -> timedelta:
        """
        :return: Packet duration as a timedelta
        """
        audio = self.get_sensors().get_audio()
        if audio is not None:
            return audio.get_duration()
        else:
            return timedelta(seconds=0)

    def get_packet_duration_s(self) -> float:
        """
        get the packet duration in seconds from the audio data
        :return: packet duration in seconds
        """
        if self.get_sensors().has_audio():
            return self.get_sensors().get_audio().get_duration_s()
        else:
            return 0.0

    def update_timestamps(self, delta_offset: float = None) -> 'WrappedRedvoxPacketM':
        """
        update all timestamps in the packet by adding the delta offset
        :param delta_offset: amount of microseconds to add to existing timestamps
        :return: WrappedRedvoxPacketM with updated timestamps
        """
        if delta_offset is None:
            delta_offset = self.get_timing_information().get_best_offset()
        redvox.api1000.common.typing.check_type(delta_offset, [float])
        # create new wrapped packet to hold changed data
        updated = WrappedRedvoxPacketM(self.get_proto())
        if self.get_proto().HasField("timing_information"):
            # update timing information timestamps
            updated.get_timing_information().set_packet_start_mach_timestamp(
                self.get_timing_information().get_packet_start_mach_timestamp() + delta_offset)
            updated.get_timing_information().set_packet_end_mach_timestamp(
                self.get_timing_information().get_packet_end_mach_timestamp() + delta_offset)
            updated.get_timing_information().set_packet_start_os_timestamp(
                self.get_timing_information().get_packet_start_os_timestamp() + delta_offset)
            updated.get_timing_information().set_packet_end_os_timestamp(
                self.get_timing_information().get_packet_end_os_timestamp() + delta_offset)
            updated.get_timing_information().set_app_start_mach_timestamp(
                self.get_timing_information().get_app_start_mach_timestamp() + delta_offset)
        if self.get_sensors().get_proto().HasField("audio"):
            # update audio first sample timestamp
            updated.get_sensors().get_audio().set_first_sample_timestamp(
                self.get_sensors().get_audio().get_first_sample_timestamp() + delta_offset)
        # todo if self.get_sensors().get_proto().HasField("compressed_audio"):
        # update compressed audio first sample timestamp
        # update timestamp payloads
        if self.get_station_information().get_station_metrics().get_proto().HasField("timestamps"):
            updated.get_station_information().get_station_metrics().get_timestamps().set_timestamps(
                self.get_station_information().get_station_metrics().get_timestamps().get_timestamps() + delta_offset,
                True)
        if self.get_sensors().get_accelerometer().get_proto().HasField("timestamps"):
            updated.get_sensors().get_accelerometer().get_timestamps().set_timestamps(
                self.get_sensors().get_accelerometer().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_ambient_temperature().get_proto().HasField("timestamps"):
            updated.get_sensors().get_ambient_temperature().get_timestamps().set_timestamps(
                self.get_sensors().get_ambient_temperature().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_gravity().get_proto().HasField("timestamps"):
            updated.get_sensors().get_gravity().get_timestamps().set_timestamps(
                self.get_sensors().get_gravity().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_gyroscope().get_proto().HasField("timestamps"):
            updated.get_sensors().get_gyroscope().get_timestamps().set_timestamps(
                self.get_sensors().get_gyroscope().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_image().get_proto().HasField("timestamps"):
            updated.get_sensors().get_image().get_timestamps().set_timestamps(
                self.get_sensors().get_image().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_light().get_proto().HasField("timestamps"):
            updated.get_sensors().get_light().get_timestamps().set_timestamps(
                self.get_sensors().get_light().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_linear_acceleration().get_proto().HasField("timestamps"):
            updated.get_sensors().get_linear_acceleration().get_timestamps().set_timestamps(
                self.get_sensors().get_linear_acceleration().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_location().get_proto().HasField("timestamps"):
            updated.get_sensors().get_location().get_timestamps().set_timestamps(
                self.get_sensors().get_location().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_magnetometer().get_proto().HasField("timestamps"):
            updated.get_sensors().get_magnetometer().get_timestamps().set_timestamps(
                self.get_sensors().get_magnetometer().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_orientation().get_proto().HasField("timestamps"):
            updated.get_sensors().get_orientation().get_timestamps().set_timestamps(
                self.get_sensors().get_orientation().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_pressure().get_proto().HasField("timestamps"):
            updated.get_sensors().get_pressure().get_timestamps().set_timestamps(
                self.get_sensors().get_pressure().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_proximity().get_proto().HasField("timestamps"):
            updated.get_sensors().get_proximity().get_timestamps().set_timestamps(
                self.get_sensors().get_proximity().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_relative_humidity().get_proto().HasField("timestamps"):
            updated.get_sensors().get_relative_humidity().get_timestamps().set_timestamps(
                self.get_sensors().get_relative_humidity().get_timestamps().get_timestamps() + delta_offset, True)
        if self.get_sensors().get_rotation_vector().get_proto().HasField("timestamps"):
            updated.get_sensors().get_rotation_vector().get_timestamps().set_timestamps(
                self.get_sensors().get_rotation_vector().get_timestamps().get_timestamps() + delta_offset, True)
        return updated


def validate_wrapped_packet(wrapped_packet: WrappedRedvoxPacketM) -> List[str]:
    """
    Validates a wrapped packet.
    :param wrapped_packet: Packet to validate.
    :return: A list of validation errors.
    """
    errors_list = _station_information.validate_station_information(wrapped_packet.get_station_information())
    errors_list.extend(_timing_information.validate_timing_information(wrapped_packet.get_timing_information()))
    errors_list.extend(_sensors.validate_sensors(wrapped_packet.get_sensors()))
    if wrapped_packet.get_api() != 1000:
        errors_list.append("Wrapped packet api is not 1000")
    return errors_list



"""
This module contains the RedVox Cloud API client.

This client provides convenient access to RedVox metadata and data. This client manages the tedious business of keeping
an up-to-date authentication token for making authenticated API requests.
"""

import contextlib
import threading
from typing import List, Optional, Tuple

import redvox.cloud.api as api
import redvox.cloud.auth_api as auth_api
import redvox.cloud.errors as cloud_errors
import redvox.common.constants as constants
import redvox.cloud.data_api as data_api
import redvox.cloud.metadata_api as metadata_api
import requests


def chunk_time_range(start_ts: int,
                     end_ts: int,
                     max_chunk: int) -> List[Tuple[int, int]]:

    if end_ts <= start_ts:
        raise cloud_errors.CloudApiError("start_ts must be < end_ts")

    if max_chunk <= 0:
        raise cloud_errors.CloudApiError("max_chunk must be > 0")

    chunks: List[Tuple[int, int]] = []

    start: int = start_ts

    while start + max_chunk < end_ts:
        chunks.append((start, start + max_chunk))
        start += max_chunk

    if start < end_ts:
        chunks.append((start, end_ts))

    return chunks


class CloudClient:
    """
    The RedVox Cloud API client.
    """

    def __init__(self,
                 username: str,
                 password: str,
                 api_conf: api.ApiConfig = api.ApiConfig.default(),
                 secret_token: Optional[str] = None,
                 refresh_token_interval: float = 600.0,
                 timeout: Optional[float] = 10.0):
        """
        Instantiates this client.
        :param username: A RedVox username.
        :param password: A RedVox password.
        :param api_conf: An optional API endpoint configuration.
        :param secret_token: An optional shared secret that may be required by the API server.
        :param refresh_token_interval: An optional interval in seconds that the auth token should be refreshed.
        """

        if refresh_token_interval <= 0:
            raise cloud_errors.CloudApiError("refresh_token_interval must be strictly > 0")

        if timeout is not None and (timeout <= 0):
            raise cloud_errors.CloudApiError("timeout must be strictly > 0")

        self.api_conf: api.ApiConfig = api_conf
        self.secret_token: Optional[str] = secret_token
        self.refresh_token_interval: float = refresh_token_interval
        self.timeout: Optional[float] = timeout

        self.__session = requests.Session()  # This must be initialized before the auth req!
        try:
            auth_resp: auth_api.AuthResp = self.authenticate_user(username, password)
        except cloud_errors.ApiConnectionError as e:
            self.close()
            raise e

        self.__refresh_timer = None
        if auth_resp.status != 200 or auth_resp.auth_token is None or len(auth_resp.auth_token) == 0:
            self.close()
            raise cloud_errors.AuthenticationError()

        self.auth_token: str = auth_resp.auth_token

        self.__refresh_timer = threading.Timer(self.refresh_token_interval, self.__refresh_token)
        self.__refresh_timer.start()

    def __refresh_token(self):
        """
        Automatically refreshes this client's auth token in the background once per minute.
        """
        try:
            self.auth_token = self.refresh_own_auth_token().auth_token
            self.__refresh_timer = threading.Timer(self.refresh_token_interval, self.__refresh_token)
            self.__refresh_timer.start()
        except:
            self.close()

    def close(self):
        """
        Terminates this client process by cancelling the refresh token timer.
        """
        try:
            if self.__refresh_timer is not None:
                self.__refresh_timer.cancel()
        except:
            pass

        try:
            if self.__session is not None:
                self.__session.close()
        except:
            pass

    def health_check(self) -> bool:
        """
        An API call that returns True if the API Cloud server is up and running or False otherwise.
        :return: True if the API Cloud server is up and running or False otherwise.
        """
        return api.health_check(self.api_conf, session=self.__session, timeout=self.timeout)

    def authenticate_user(self, username: str, password: str) -> auth_api.AuthResp:
        """
        Attempts to authenticate the given RedVox user.
        :param username: The RedVox username.
        :param password: The RedVox password.
        :return: An authenticate response.
        """
        if len(username) == 0:
            raise cloud_errors.CloudApiError("Username must be provided")

        if len(password) == 0:
            raise cloud_errors.CloudApiError("Password must be provided")

        auth_req: auth_api.AuthReq = auth_api.AuthReq(username, password)
        return auth_api.authenticate_user(self.api_conf,
                                          auth_req,
                                          session=self.__session,
                                          timeout=self.timeout)

    def validate_auth_token(self, auth_token: str) -> Optional[auth_api.ValidateTokenResp]:
        """
        Validates the provided authentication token with the cloud API.
        :param auth_token: Authentication token to validate.
        :return: An authentication response with token details or None if token in invalid
        """
        if len(auth_token) == 0:
            raise cloud_errors.CloudApiError("auth_token must be provided")

        token_req: auth_api.ValidateTokenReq = auth_api.ValidateTokenReq(auth_token)
        return auth_api.validate_token(self.api_conf, token_req, session=self.__session, timeout=self.timeout)

    def validate_own_auth_token(self) -> Optional[auth_api.ValidateTokenResp]:
        """
        Validates the current token held by this client.
        :return: An authentication response with token details or None if token in invalid
        """
        return self.validate_auth_token(self.auth_token)

    def refresh_auth_token(self, auth_token: str) -> Optional[auth_api.RefreshTokenResp]:
        """
        Retrieves a new authentication token from a given valid authentication token.
        :param auth_token: The authentication token to verify.
        :return: A new authentication token or None if the provide auth token is not valid.
        """
        if len(auth_token) == 0:
            raise cloud_errors.CloudApiError("auth_token must be provided")

        refresh_token_req: auth_api.RefreshTokenReq = auth_api.RefreshTokenReq(auth_token)
        return auth_api.refresh_token(self.api_conf,
                                      refresh_token_req,
                                      session=self.__session,
                                      timeout=self.timeout)

    def refresh_own_auth_token(self) -> Optional[auth_api.RefreshTokenResp]:
        """
        Retrieves a new authentication token using the token held by this client as a reference.
        :return: A new authentication token or None if the provide auth token is not valid.
        """
        return self.refresh_auth_token(self.auth_token)

    def request_metadata(self,
                         start_ts_s: int,
                         end_ts_s: int,
                         station_ids: List[str],
                         metadata_to_include: List[str],
                         chunk_by_seconds: int = constants.SECONDS_PER_DAY) -> Optional[metadata_api.MetadataResp]:
        """
        Requests RedVox packet metadata.
        :param start_ts_s: Start epoch of request window.
        :param end_ts_s: End epoch of request window.
        :param station_ids: A list of station ids.
        :param metadata_to_include: A list of metadata fields to include (see: redvox.cloud.metadata.AvailableMetadata)
        :param chunk_by_seconds: Split up longer requests into chunks of chunk_by_seconds size (default 86400s/1d)
        :return: A metadata result containing the requested metadata or None on error.
        """
        if end_ts_s <= start_ts_s:
            raise cloud_errors.CloudApiError("start_ts_s must be < end_ts_s")

        if len(station_ids) == 0:
            raise cloud_errors.CloudApiError("At least one station_id must be included")

        if len(metadata_to_include) == 0:
            raise cloud_errors.CloudApiError("At least one metadata field must be included")

        if chunk_by_seconds <= 0:
            raise cloud_errors.CloudApiError("chunk_by_seconds must be > 0")

        time_chunks: List[Tuple[int, int]] = chunk_time_range(start_ts_s, end_ts_s, chunk_by_seconds)
        metadata_resp: metadata_api.MetadataResp = metadata_api.MetadataResp([])

        for start_ts, end_ts in time_chunks:
            metadata_req: metadata_api.MetadataReq = metadata_api.MetadataReq(self.auth_token,
                                                                              start_ts,
                                                                              end_ts,
                                                                              station_ids,
                                                                              metadata_to_include,
                                                                              self.secret_token)

            chunked_resp: Optional[metadata_api.MetadataResp] = metadata_api.request_metadata(self.api_conf,
                                                                                              metadata_req,
                                                                                              session=self.__session,
                                                                                              timeout=self.timeout)

            if chunked_resp:
                metadata_resp.metadata.extend(chunked_resp.metadata)

        return metadata_resp

    def request_metadata_m(self,
                           start_ts_s: int,
                           end_ts_s: int,
                           station_ids: List[str],
                           metadata_to_include: List[str],
                           chunk_by_seconds: int = constants.SECONDS_PER_DAY) -> Optional[metadata_api.MetadataRespM]:
        """
        Requests RedVox packet metadata.
        :param start_ts_s: Start epoch of request window.
        :param end_ts_s: End epoch of request window.
        :param station_ids: A list of station ids.
        :param metadata_to_include: A list of metadata fields to include (see: redvox.cloud.metadata.AvailableMetadata)
        :param chunk_by_seconds: Split up longer requests into chunks of chunk_by_seconds size (default 86400s/1d)
        :return: A metadata result containing the requested metadata or None on error.
        """
        if end_ts_s <= start_ts_s:
            raise cloud_errors.CloudApiError("start_ts_s must be < end_ts_s")

        if len(station_ids) == 0:
            raise cloud_errors.CloudApiError("At least one station_id must be included")

        if len(metadata_to_include) == 0:
            raise cloud_errors.CloudApiError("At least one metadata field must be included")

        if chunk_by_seconds <= 0:
            raise cloud_errors.CloudApiError("chunk_by_seconds must be > 0")

        time_chunks: List[Tuple[int, int]] = chunk_time_range(start_ts_s, end_ts_s, chunk_by_seconds)
        metadata_resp: metadata_api.MetadataRespM = metadata_api.MetadataRespM([])

        for start_ts, end_ts in time_chunks:
            metadata_req: metadata_api.MetadataReq = metadata_api.MetadataReq(self.auth_token,
                                                                              start_ts,
                                                                              end_ts,
                                                                              station_ids,
                                                                              metadata_to_include,
                                                                              self.secret_token)

            chunked_resp: Optional[metadata_api.MetadataRespM] = metadata_api.request_metadata_m(self.api_conf,
                                                                                                 metadata_req,
                                                                                                 session=self.__session,
                                                                                                 timeout=self.timeout)

            if chunked_resp:
                metadata_resp.db_packets.extend(chunked_resp.db_packets)

        return metadata_resp

    def request_timing_metadata(self,
                                start_ts_s: int,
                                end_ts_s: int,
                                station_ids: List[str],
                                chunk_by_seconds: int = constants.SECONDS_PER_DAY) -> metadata_api.TimingMetaResponse:
        """
        Requests timing metadata from RedVox packets.
        :param start_ts_s: Start epoch of the request.
        :param end_ts_s: End epoch of the request.
        :param station_ids: A list of station ids.
        :param chunk_by_seconds: Split up longer requests into chunks of chunk_by_seconds size (default 86400s/1d)
        :return: A response containing the requested metadata.
        """
        if end_ts_s <= start_ts_s:
            raise cloud_errors.CloudApiError("start_ts_s must be < end_ts_s")

        if len(station_ids) == 0:
            raise cloud_errors.CloudApiError("At least one station_id must be included")

        if chunk_by_seconds <= 0:
            raise cloud_errors.CloudApiError("chunk_by_seconds must be > 0")

        time_chunks: List[Tuple[int, int]] = chunk_time_range(start_ts_s, end_ts_s, chunk_by_seconds)
        metadata_resp: metadata_api.TimingMetaResponse = metadata_api.TimingMetaResponse([])

        for start_ts, end_ts in time_chunks:
            timing_req: metadata_api.TimingMetaRequest = metadata_api.TimingMetaRequest(self.auth_token,
                                                                                        start_ts,
                                                                                        end_ts,
                                                                                        station_ids,
                                                                                        self.secret_token)
            chunked_resp: metadata_api.TimingMetaResponse = metadata_api.request_timing_metadata(self.api_conf,
                                                                                                 timing_req,
                                                                                                 session=self.__session,
                                                                                                 timeout=self.timeout)

            if chunked_resp:
                metadata_resp.items.extend(chunked_resp.items)

        return metadata_resp

    def request_report_data(self, report_id: str) -> Optional[data_api.ReportDataResp]:
        """
        Requests a signed URL for a given report ID.
        :param report_id: The report ID to request data for.
        :return: A response containing a signed URL of the report data.
        """
        if len(report_id) == 0:
            raise cloud_errors.CloudApiError("report_id must be included")

        report_data_req: data_api.ReportDataReq = data_api.ReportDataReq(self.auth_token,
                                                                         report_id,
                                                                         self.secret_token)
        return data_api.request_report_data(self.api_conf, report_data_req, session=self.__session, timeout=self.timeout)

    def request_data_range(self,
                           start_ts_s: int,
                           end_ts_s: int,
                           station_ids: List[str]) -> data_api.DataRangeResp:
        """
        Request signed URLs for RedVox packets.
        :param start_ts_s: The start epoch of the window.
        :param end_ts_s:  The end epoch of the window.
        :param station_ids: A list of station ids.
        :return: A response containing a list of signed URLs for the RedVox packets.
        """
        if end_ts_s <= start_ts_s:
            raise cloud_errors.CloudApiError("start_ts_s must be < end_ts_s")

        if len(station_ids) == 0:
            raise cloud_errors.CloudApiError("At least one station_id must be provided")

        data_range_req: data_api.DataRangeReq = data_api.DataRangeReq(self.auth_token,
                                                                      start_ts_s,
                                                                      end_ts_s,
                                                                      station_ids,
                                                                      self.secret_token)

        return data_api.request_range_data(self.api_conf, data_range_req, session=self.__session, timeout=self.timeout)


@contextlib.contextmanager
def cloud_client(username: str,
                 password: str,
                 api_conf: api.ApiConfig = api.ApiConfig.default(),
                 secret_token: Optional[str] = None,
                 refresh_token_interval: float = 600.0,
                 timeout: float = 10.0):
    """
    Function that can be used within a "with" block to automatically handle the closing of open resources.
    Creates and returns a CloudClient that will automatically be closed when exiting the with block or if an error
    occurs.

    See https://docs.python.org/3/library/contextlib.html for more info.

    :param username: The Cloud API username.
    :param password: The Cloud API password.
    :param api_conf: The Cloud API endpoint configuration.
    :param secret_token: An optional secret token.
    :param refresh_token_interval: An optional token refresh interval
    :return: A CloudClient.
    """
    client: CloudClient = CloudClient(username, password, api_conf, secret_token, refresh_token_interval, timeout)
    try:
        yield client
    finally:
        if client is not None:
            client.close()




"""
This module encapsulates audio sensor metadata and data.
"""

import enum
from datetime import timedelta

import numpy as np
import redvox.api1000.common.typing
import redvox.api1000.proto.redvox_api_m_pb2 as redvox_api_m_pb2
import redvox.api1000.common.common as common
import redvox.api1000.common.generic
from redvox.api1000.wrapped_redvox_packet.station_information import AudioSamplingRate
from redvox.api1000.common.decorators import wrap_enum

from typing import List


class Audio(redvox.api1000.common.generic.ProtoBase[redvox_api_m_pb2.RedvoxPacketM.Sensors.Audio]):
    """
    This class encapsulates audio metadata and data.
    """
    def __init__(self, proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.Audio):
        super().__init__(proto)
        self._samples: common.SamplePayload = common.SamplePayload(proto.samples)

    @staticmethod
    def new() -> 'Audio':
        """
        :return: A new, empty Audio sensor instance.
        """
        proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.Audio \
            = redvox_api_m_pb2.RedvoxPacketM.Sensors.Audio()
        return Audio(proto)

    def get_sensor_description(self) -> str:
        """
        :return: This sensor's description.
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> 'Audio':
        """
        Sets this sensor's description.
        :param sensor_description: Description to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_first_sample_timestamp(self) -> float:
        """
        :return: The machine timestamp associated with the first audio sample.
        """
        return self._proto.first_sample_timestamp

    def set_first_sample_timestamp(self, first_sample_timestamp: float) -> 'Audio':
        """
        Sets the machine timestamp that corresponds to the first audio sample.
        :param first_sample_timestamp: Timestamp to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(first_sample_timestamp, [float, int])
        self._proto.first_sample_timestamp = first_sample_timestamp
        return self

    def get_sample_rate(self) -> float:
        """
        :return: The sampling rate in Hz
        """
        return self._proto.sample_rate

    def set_sample_rate(self, sample_rate: float) -> 'Audio':
        """
        Sets the sampling rate.
        :param sample_rate: Sampling rate Hz.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sample_rate, [int, float])
        self._proto.sample_rate = sample_rate
        return self

    def get_bits_of_precision(self) -> float:
        """
        :return: Returns the bits of precision or dynamic range of the audio samples.
        """
        return self._proto.bits_of_precision

    def set_bits_of_precision(self, bits_of_precision: float) -> 'Audio':
        """
        Set the bits or precision (dynamic range) of the audio samples.
        :param bits_of_precision: Bits of precision
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(bits_of_precision, [int, float])
        self._proto.bits_of_precision = bits_of_precision
        return self

    def get_is_scrambled(self) -> bool:
        """
        :return: If this audio has been scrambled to remove potential voice data.
        """
        return self._proto.is_scrambled

    def set_is_scrambled(self, is_scrambled: bool) -> 'Audio':
        """
        Sets if the audio has been scrambled or not.
        :param is_scrambled: True if scrambled, False otherwise
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(is_scrambled, [bool])
        self._proto.is_scrambled = is_scrambled
        return self

    def get_encoding(self) -> str:
        """
        :return: The audio encoding scheme as defined by the app.
        """
        return self._proto.encoding

    def set_encoding(self, encoding: str) -> 'Audio':
        """
        Sets the audio encoding scheme.
        :param encoding: Encoding to use.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(encoding, [str])
        self._proto.encoding = encoding
        return self

    def get_samples(self) -> common.SamplePayload:
        """
        :return: The sample payload associated with the audio sensor.
        """
        return self._samples

    def set_samples(self, audio_samples: common.SamplePayload) -> 'Audio':
        """
        Sets the SamplePayload associated with the audio sensor.
        :param audio_samples:
        :return:
        """
        self._samples.set_unit(audio_samples.get_unit())
        self._samples.set_values(audio_samples.get_values())
        return self

    def get_num_samples(self) -> int:
        """
        :return: The number of audio samples stored in this packet.
        """
        return self.get_samples().get_values_count()

    def get_duration(self) -> timedelta:
        """
        :return: Duration of this packet represented as a timedelta
        """
        return timedelta(seconds=self.get_duration_s())

    def get_duration_s(self) -> float:
        """
        calculate the duration of the audio data in seconds
        :return: duration of audio data in seconds
        """
        return float(self.get_num_samples()) / self.get_sample_rate()


def validate_audio(audio_sensor: Audio) -> List[str]:
    """
    Validates the audio sensor.
    :param audio_sensor: Audio sensor to validate.
    :return: A list of validation errors.
    """
    # todo: add default audio unit, normalization factor if needed
    errors_list = common.validate_sample_payload(audio_sensor.get_samples(), "Audio")
    if audio_sensor.get_samples().get_values_count() > 0:
        if np.min(audio_sensor.get_samples().get_values()) < -1.0:
            errors_list.append("Audio minimum value of samples cannot be less than -1.0")
        if np.max(audio_sensor.get_samples().get_values()) > 1.0:
            errors_list.append("Audio maximum value of samples cannot be greater than 1.0")
    if audio_sensor.get_first_sample_timestamp() == 0:
        errors_list.append("Audio first sample timestamp is default value")
    if AudioSamplingRate.from_sampling_rate(audio_sensor.get_sample_rate()) is None:
        errors_list.append("Audio sample rate is not a valid value")
    return errors_list


# noinspection Mypy
@wrap_enum(redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio.AudioCodec)
class AudioCodec(enum.Enum):
    """
    Audio codec used for compressed audio streams
    """
    UNKNOWN: int = 0
    FLAC: int = 1


class CompressedAudio(
        redvox.api1000.common.generic.ProtoBase[redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio]):
    """
    Encapsulates metadata and data for compressed audio streams.
    """
    def __init__(self, proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio):
        super().__init__(proto)

    @staticmethod
    def new() -> 'CompressedAudio':
        """
        :return: A new, empty CompressedAudio sensor instance
        """
        proto: redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio \
            = redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio()
        return CompressedAudio(proto)

    def get_sensor_description(self) -> str:
        """
        :return: This sensor's description.
        """
        return self._proto.sensor_description

    def set_sensor_description(self, sensor_description: str) -> 'CompressedAudio':
        """
        Sets this sensor's description.
        :param sensor_description: Description to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sensor_description, [str])
        self._proto.sensor_description = sensor_description
        return self

    def get_first_sample_timestamp(self) -> float:
        """
        :return: The machine timestamp associated with the first audio sample.
        """
        return self._proto.first_sample_timestamp

    def set_first_sample_timestamp(self, first_sample_timestamp: float) -> 'CompressedAudio':
        """
        Sets the machine timestamp that corresponds to the first audio sample.
        :param first_sample_timestamp: Timestamp to set.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(first_sample_timestamp, [float, int])
        self._proto.first_sample_timestamp = first_sample_timestamp
        return self

    def get_sample_rate(self) -> float:
        """
        :return: The sampling rate in Hz
        """
        return self._proto.sample_rate

    def set_sample_rate(self, sample_rate: float) -> 'CompressedAudio':
        """
        Sets the sampling rate.
        :param sample_rate: Sampling rate Hz.
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(sample_rate, [int, float])
        self._proto.sample_rate = sample_rate
        return self

    def get_is_scrambled(self) -> bool:
        """
        :return: If this audio has been scrambled to remove potential voice data.
        """
        return self._proto.is_scrambled

    def set_is_scrambled(self, is_scrambled: bool) -> 'CompressedAudio':
        """
        Sets if the audio has been scrambled or not.
        :param is_scrambled: True if scrambled, False otherwise
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(is_scrambled, [bool])
        self._proto.is_scrambled = is_scrambled
        return self

    def get_audio_bytes(self) -> bytes:
        """
        :return: The compressed bytes that make up the audio stream
        """
        return self._proto.audio_bytes

    def set_audio_bytes(self, audio_bytes: bytes) -> 'CompressedAudio':
        """
        Set the bytes that make up the compressed audio stream
        :param audio_bytes:
        :return:
        """
        redvox.api1000.common.typing.check_type(audio_bytes, [bytes])
        self._proto.audio_bytes = audio_bytes
        return self

    def get_audio_codec(self) -> AudioCodec:
        """
        :return: The audio codec used to compress the audio
        """
        # noinspection Mypy
        return self._proto.audio_codec

    def set_audio_codec(self, audio_codec: AudioCodec) -> 'CompressedAudio':
        """
        Sets the audio codec used to compress this audio stream
        :param audio_codec: Audio codec used to compress this stream
        :return: A modified instance of self
        """
        redvox.api1000.common.typing.check_type(audio_codec, [AudioCodec])
        self.get_proto().audio_codec = \
            redvox_api_m_pb2.RedvoxPacketM.Sensors.CompressedAudio.AudioCodec.Value(audio_codec.name)
        return self


def validate_compress_audio(compress_audio: CompressedAudio) -> List[str]:
    """
    Validates compressed audio.
    :param compress_audio: CompressedAudio to validate.
    :return: A list of validation errors.
    """
    errors_list = []
    if len(compress_audio.get_audio_bytes()) < 1:
        errors_list.append("Compressed Audio sample audio bytes is empty.")
    if compress_audio.get_first_sample_timestamp() == 0:
        errors_list.append("Compressed Audio first sample timestamp is default value")
    if AudioSamplingRate.from_sampling_rate(compress_audio.get_sample_rate()) is None:
        errors_list.append("Compressed Audio sample rate is not a valid value")
    return errors_list



"""
This module creates specific time-bounded segments of data for users
"""
import pandas as pd
import numpy as np
from typing import Optional, Set
from dataclasses import dataclass
from redvox.common import date_time_utils as dtu
from redvox.common.sensor_data import SensorType
from redvox.common.load_sensor_data import ReadResult, read_all_in_dir


DEFAULT_GAP_TIME_S: float = 0.25
DEFAULT_START_PADDING_S: float = 120.
DEFAULT_END_PADDING_S: float = 120.


@dataclass
class DataWindow:
    """
    Holds the data for a given time window; adds interpolated timestamps to fill gaps and pad start and end values
    Properties:
        input_directory: string, directory that contains the files to read data from.  REQUIRED
        station_ids: optional set of strings, list of station ids to filter on.
                        If empty or None, get any ids found in the input directory.  Default None
        start_datetime: optional datetime, start datetime of the window.
                        If None, uses the first timestamp of the filtered data.  Default None
        end_datetime: optional datetime, end datetime of the window.
                        If None, uses the last timestamp of the filtered data.  Default None
        start_padding_s: float, the amount of seconds to include before the start_datetime
                            when filtering data.  Default DEFAULT_START_PADDING_S
        end_padding_s: float, the amount of seconds to include after the end_datetime
                        when filtering data.  Default DEFAULT_END_PADDING_S
        gap_time_s: float, the minimum amount of seconds between data points that would indicate a gap.
                    Default DEFAULT_GAP_TIME_S
        apply_correction: bool, if True, update the timestamps in the data based on best station offset.  Default False
        structured_layout: bool, if True, the input_directory contains specially named and organized
                            directories of data.  Default False
        stations: optional ReadResult, the results of reading the data from input_directory
    """
    input_directory: str
    station_ids: Optional[Set[str]] = None
    start_datetime: Optional[dtu.datetime] = None
    end_datetime: Optional[dtu.datetime] = None
    start_padding_s: float = DEFAULT_START_PADDING_S
    end_padding_s: float = DEFAULT_END_PADDING_S
    gap_time_s: float = DEFAULT_GAP_TIME_S
    apply_correction: bool = False
    structured_layout: bool = False
    stations: Optional[ReadResult] = None

    def __post_init__(self):
        """
        loads the data after initialization
        """
        if self.stations is None:
            self.read_data()

    def copy(self) -> 'DataWindow':
        """
        :return: a copy of the DataWindow
        """
        return DataWindow(self.input_directory, self.station_ids, self.start_datetime, self.end_datetime,
                          self.start_padding_s, self.end_padding_s, self.gap_time_s,
                          self.apply_correction, self.structured_layout, self.stations)

    def _has_time_window(self) -> bool:
        """
        Returns true if there is a start or end datetime in the settings
        :return: True if start_datetime or end_datetime exists
        """
        return self.start_datetime is not None or self.end_datetime is not None

    def _pad_start_datetime_s(self) -> float:
        """
        apply padding to the start datetime
        :return: padded start datetime as seconds since epoch UTC
        """
        return dtu.datetime_to_epoch_seconds_utc(self.start_datetime) - self.start_padding_s

    def _pad_end_datetime_s(self) -> float:
        """
        apply padding to the end datetime
        :return: padded end datetime as seconds since epoch UTC
        """
        return dtu.datetime_to_epoch_seconds_utc(self.end_datetime) + self.end_padding_s

    def correct_timestamps(self):
        """
        update the timestamps in all stations
        """
        for station in self.stations.station_id_uuid_to_stations.values():
            if self.apply_correction:
                station.update_timestamps()

    def data_padder(self, data_df: pd.DataFrame, sample_interval_s: float) -> pd.DataFrame:
        """
        Pad the start and end of the dataframe with np.nan
        :param data_df: dataframe with timestamps as column "timestamps"
        :param sample_interval_s: constant sample interval in seconds
        :return: dataframe padded with np.nans in front and back to meet full size of expected start and end
        """
        # extract the necessary information to pad the data
        data_time_stamps = data_df.sort_values("timestamps")["timestamps"].to_numpy()
        first_data_timestamp = data_time_stamps[0]
        last_data_timestamp = data_time_stamps[-1]
        expected_start = dtu.datetime_to_epoch_microseconds_utc(self.start_datetime)
        expected_end = dtu.datetime_to_epoch_microseconds_utc(self.end_datetime)
        result_df = data_df.copy()
        # FRONT/END GAP FILL!  calculate the audio samples missing based on inputs
        if expected_start < first_data_timestamp:
            start_diff = first_data_timestamp - expected_start
            num_missing_samples = int(dtu.microseconds_to_seconds(start_diff) / sample_interval_s) + 1
            # add the gap data to the result dataframe
            result_df = result_df.append(create_empty_df(expected_start -
                                                         dtu.seconds_to_microseconds(sample_interval_s),
                                                         sample_interval_s, data_df.columns,
                                                         num_missing_samples), ignore_index=True)
        if expected_end > last_data_timestamp:
            last_diff = expected_end - last_data_timestamp
            num_missing_samples = int(dtu.microseconds_to_seconds(last_diff) / sample_interval_s) + 1
            # add the gap data to the result dataframe
            result_df = result_df.append(create_empty_df(expected_end +
                                                         dtu.seconds_to_microseconds(sample_interval_s),
                                                         sample_interval_s, data_df.columns,
                                                         num_missing_samples, True), ignore_index=True)
        return result_df.sort_values("timestamps", ignore_index=True)

    def read_data(self):
        """
        read data using the properties of the class
        """
        start_time = int(self._pad_start_datetime_s()) if self.start_datetime else None
        end_time = int(self._pad_end_datetime_s()) if self.end_datetime else None
        self.stations = read_all_in_dir(self.input_directory, start_time, end_time,
                                        self.station_ids, self.structured_layout)

        ids_to_pop = []
        # check if ids in station data from files
        for ids in self.station_ids:
            if not self.stations.check_for_id(ids):
                # error handling
                print(f"WARNING: {ids} doesn't have any data to read")
            elif not self.stations.get_station(ids).has_audio_data():
                # no audio data is about the same as no data
                print(f"WARNING: {ids} doesn't have any audio data to read")
                ids_to_pop.append(ids)
        for ids in ids_to_pop:
            self.stations.pop_station(ids)
        # calculate time differences in audio samples
        for station in self.stations.get_all_stations():
            # apply time correction
            if self.apply_correction:
                station.update_timestamps()
            for packet in range(len(station.packet_data) - 1):
                data_start = station.packet_data[packet].data_start_timestamp
                data_num_samples = station.packet_data[packet].packet_num_audio_samples
                next_packet_start_index = \
                    station.audio_sensor().data_df.query("timestamps == @data_start").first_valid_index() + \
                    data_num_samples
                data_end = station.audio_sensor().data_timestamps()[next_packet_start_index - 1]
                next_packet_start = station.audio_sensor().data_timestamps()[next_packet_start_index]
                if next_packet_start - data_end < dtu.seconds_to_microseconds(self.gap_time_s):
                    station.packet_data[packet].sample_interval_to_next_packet = \
                        (next_packet_start - data_start) / data_num_samples

    def create_window(self) -> 'DataWindow':
        """
        constrain the data to the window specified by the parameters
        :return: only the data in the window specified by the parameters
        """
        if self._has_time_window():
            new_data_window = self.copy()
            ids_to_pop = []
            # fill in gaps and truncate
            for station_id, station in new_data_window.stations.station_id_uuid_to_stations.items():
                # prepare a bunch of information to be used later
                if new_data_window.start_datetime:
                    start_timestamp = dtu.seconds_to_microseconds(
                        dtu.datetime_to_epoch_seconds_utc(new_data_window.start_datetime))
                else:
                    start_timestamp = station.audio_sensor().first_data_timestamp()
                if new_data_window.end_datetime:
                    end_timestamp = dtu.seconds_to_microseconds(
                        dtu.datetime_to_epoch_seconds_utc(new_data_window.end_datetime))
                else:
                    end_timestamp = station.audio_sensor().last_data_timestamp()
                # truncate packets to include only the ones with the data for the window
                station.packet_data = [p for p in station.packet_data
                                       if p.data_end_timestamp > start_timestamp and
                                       p.data_start_timestamp < end_timestamp]
                for sensor_type, sensor in station.station_data.items():
                    # TRUNCATE!  get only the timestamps between the start and end timestamps
                    df_timestamps = sensor.data_timestamps()
                    if len(df_timestamps) < 1:
                        print(f"WARNING: Data window for {station.station_metadata.station_id} {sensor_type.name} "
                              f"sensor has no data points!")
                        if sensor_type == SensorType.AUDIO:
                            ids_to_pop.append(station.station_metadata.station_id)
                        break
                    temp = np.where(
                        (start_timestamp < df_timestamps) & (df_timestamps < end_timestamp))[0]
                    # oops, all the samples have been cut off
                    if len(temp) < 1:
                        print(f"WARNING: Data window for {station.station_metadata.station_id} {sensor_type.name} "
                              f"sensor has truncated all data points")
                        if sensor_type == SensorType.LOCATION:
                            # take the locations before the start_timestamp as valid locations
                            temp = np.where(df_timestamps < end_timestamp)[0]
                            if len(temp) < 1:
                                break
                            else:
                                print(f"Using all {sensor_type.name} data points before {end_timestamp} instead")
                        else:
                            if sensor_type == SensorType.AUDIO:
                                ids_to_pop.append(station.station_metadata.station_id)
                            break
                    sensor.data_df = sensor.data_df.iloc[temp].reset_index(drop=True)
                    if sensor.is_sample_interval_invalid():
                        print(f"WARNING: {sensor_type.name} has undefined sample interval and sample rate!")
                        break
                    # GAP FILL
                    sensor.data_df = gap_filler(sensor.data_df, sensor.sample_interval_s)
                    # PAD DATA
                    sensor.data_df = new_data_window.data_padder(sensor.data_df, sensor.sample_interval_s)
            # remove any station without audio sensor
            for ids in ids_to_pop:
                new_data_window.stations.pop_station(ids)
            return new_data_window
        else:
            return self


def gap_filler(data_df: pd.DataFrame, sample_interval_s: float,
               gap_duration_s: float = DEFAULT_GAP_TIME_S) -> pd.DataFrame:
    """
    fills gaps in the dataframe with np.nan by interpolating timestamps based on the mean expected sample interval
    :param data_df: dataframe with timestamps as column "timestamps"
    :param sample_interval_s: sample interval in seconds
    :param gap_duration_s: duration in seconds of minimum missing data to be considered a gap
    :return: dataframe without gaps
    """
    # extract the necessary information to compute gap size and gap timestamps
    data_time_stamps = data_df.sort_values("timestamps", ignore_index=True)["timestamps"].to_numpy()
    first_data_timestamp = data_time_stamps[0]
    last_data_timestamp = data_time_stamps[-1]
    data_duration_s = dtu.microseconds_to_seconds(last_data_timestamp - first_data_timestamp)
    num_points = len(data_time_stamps)
    # add one to calculation to include the last timestamp
    expected_num_points = int(data_duration_s / sample_interval_s) + 1
    # gap duration cannot be less than sample interval
    if gap_duration_s < sample_interval_s:
        gap_duration_s = sample_interval_s
    result_df = data_df.copy()
    # if there are less points than our expected amount, we have gaps to fill
    if num_points < expected_num_points:
        # if the data we're looking at is short enough, we can start comparing points
        if num_points < 1000:
            # look at every timestamp except the last one
            for index in range(0, num_points - 1):
                # compare that timestamp to the next
                time_diff = dtu.microseconds_to_seconds(data_time_stamps[index + 1] - data_time_stamps[index])
                # calc samples to add, subtracting 1 to prevent copying last timestamp
                num_new_samples = int(time_diff / sample_interval_s) - 1
                if time_diff > gap_duration_s and num_new_samples > 0:
                    # add the gap data to the result dataframe
                    result_df = result_df.append(create_empty_df(data_time_stamps[index], sample_interval_s,
                                                                 data_df.columns, num_new_samples), ignore_index=True)
        else:
            # too many points to check, divide and conquer using recursion!
            half_samples = int(num_points / 2)
            first_data_df = data_df.iloc[:half_samples].copy().reset_index(drop=True)
            second_data_df = data_df.iloc[half_samples:].copy().reset_index(drop=True)
            # give half the samples to each recursive call
            first_data_df = gap_filler(first_data_df, sample_interval_s, gap_duration_s)
            second_data_df = gap_filler(second_data_df, sample_interval_s, gap_duration_s)
            result_df = first_data_df.append(second_data_df, ignore_index=True)
    return result_df.sort_values("timestamps", ignore_index=True)


def create_empty_df(start_timestamp: float, sample_interval_s: float, columns: pd.Index,
                    num_samples_to_add: int, add_to_start: bool = False) -> pd.DataFrame:
    """
    Creates an empty dataframe with num_samples_to_add - 1 timestamps, using columns as the columns
    The one timestamp not being added would be a copy of the start timestamp.
    :param start_timestamp: timestamp to start calculating other timestamps from
    :param sample_interval_s: fixed sample interval in seconds
    :param columns: the non-timestamp columns of the dataframe
    :param num_samples_to_add: the number of timestamps to create
    :param add_to_start: if True, subtracts sample_interval_s from start_timestamp, default False
    :return:
    """
    if add_to_start:
        sample_interval_s = -sample_interval_s
    new_timestamps = np.vectorize(lambda t: start_timestamp + dtu.seconds_to_microseconds(t * sample_interval_s))(
        list(range(1, num_samples_to_add + 1)))
    empty_df = pd.DataFrame([], columns=columns)
    for column_index in columns:
        if column_index == "timestamps":
            empty_df["timestamps"] = new_timestamps
        else:
            empty_df[column_index] = np.nan
    # return a dataframe with only timestamps
    return empty_df
